##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from struct import pack, unpack
from typing import Any, Callable, Final, Union

from numpy import float64, uint32, uint64
from randomgen import UserBitGenerator

from .caches import BasicCache as BitCache
from .platforms import QuantumPlatform
from .protocols import QuantumProtocol


###############################################################################
## QUANTUM BIT GENERATOR (FACADE)
###############################################################################
class QuantumBitGenerator(UserBitGenerator):
    def __init__(
        self,
        platform: QuantumPlatform,
        protocol: QuantumProtocol,
        ISRAW32: bool = False,
    ) -> None:
        self.platform: QuantumPlatform = platform
        self.protocol: QuantumProtocol = protocol
        self._ISRAW32: Final[bool] = ISRAW32
        self._bitcache: BitCache = BitCache()
        super().__init__(
            bits=self.BITS,
            next_raw=self._next_raw,
            next_32=self._next_32,
            next_64=self._next_64,
            next_double=self._next_double,
        )

    ############################### PUBLIC API ###############################
    @property
    def BITS(self) -> int:
        """
        Either 32 or 64. The number of bits output by NumPy's `random_raw()`
        method. Final, it cannot be modified after instantiation through the
        ISRAW32 parameter.
        """
        return 32 if self._ISRAW32 else 64

    @property
    def platform(self) -> QuantumPlatform:
        return self._platform

    @platform.setter
    def platform(self, p: QuantumPlatform) -> None:
        self._platform = p

    @property
    def protocol(self) -> QuantumProtocol:
        return self._protocol

    @protocol.setter
    def protocol(self, p: QuantumProtocol) -> None:
        self._protocol = p

    def dump_cache(self, flush: bool = False) -> str:
        """
        Returns all the contents stored in the cache.

        PARAMETERS
        ----------
        flush: bool
            If `True` erase the cache after dumping.

        RETURNS
        -------
        out: str
            The bitstring stored in cache.
        """
        bitstring: str = self._bitcache.dump()
        if flush:
            self._bitcache.flush()
        return bitstring

    def flush_cache(self) -> None:
        """
        Erase the cache.

        RETURNS
        -------
        out: bool
            `True` if succeeds, `False` otherwise.
        """
        self._bitcache.flush()

    def load_cache(self, bitstring: str, flush: bool = False) -> None:
        """
        Load cache contents from bitstring.

        PARAMETERS
        ----------
        bitstring: str
            The bitstring to load to cache.
        flush: bool
            If `True` erase cache before loading.

        RETURNS
        -------
        out: bool
            `True` if succeeds, `False` otherwise.

        RAISES
        ------
        TypeError (push)
            If input bitstring is not str
        ValueError (push)
            If input bitstring is not a valid bitstring
        """
        if flush:
            self._bitcache.flush()
        self._bitcache.push(bitstring)

    def random_bitstring(self, num_bits: int = 0) -> str:
        """
        Returns a random bitstring of a given lenght. If less than one it
        defaults to the raw number of bits for the instance QiskitBitGenerator
        (i.e. 32 or 64).

        PARAMETERS
        ----------
        num_bits: int
            Number of bits to retrieve.

        RETURNS
        -------
        out: str
            Bitstring of lenght `num_bits`.
        """
        if num_bits < 1:
            num_bits = self.BITS
        while self._bitcache.size < num_bits:
            self._refill_cache()
        return self._bitcache.pop(num_bits)

    def random_double(self, n: float = 1) -> float:
        """
        Returns a random double from a uniform distribution in the range
        [0,n). Defaults to [0,1).

        PARAMETERS
        ----------
        n: float
            Size of the range [0,n) from which to draw the random number.

        RETURNS
        -------
        out: float
            Random float in the range [0,n).

        COPYRIGHT NOTICE
        ----------------
        Source: https://github.com/ozanerhansha/qRNG
        License: GNU GENERAL PUBLIC LICENSE VERSION 3
        Changes:
            - Add static type hints
            - Limit range to [0,n) instead of [min,max) and add default
            - Replace call to original get_random_int64
        """
        unpacked = 0x3FF0000000000000 | self.random_uint(64) >> 12
        packed = pack("Q", unpacked)
        value: float = unpack("d", packed)[0] - 1.0
        return value * n

    def random_uint(self, num_bits: int = 0) -> int:
        """
        Returns a random unsigned int of a given size in bits.

        PARAMETERS
        ----------
        num_bits: int
            Number of bits to retrieve. If less than one it defaults to the raw
            number of bits for the instance QiskitBitGenerator (i.e. 32 or 64).

        RETURNS
        -------
        out: int
            Unsigned int of `num_bits` bits.
        """
        if num_bits < 1:
            num_bits = self.BITS
        return int(self.random_bitstring(num_bits), 2)

    ############################### PRIVATE API ###############################
    def _refill_cache(self) -> None:
        bitstring: str = self.platform.fetch_random_bits(self.protocol)
        self._bitcache.push(bitstring)

    ############################# NUMPY INTERFACE #############################
    @property
    def _next_raw(self) -> Callable[[Any], Union[uint32, uint64]]:
        """
        A callable that returns either 64 or 32 random bits. It must accept
        a single input which is a void pointer to a memory address.
        """
        return self._next_32 if self._ISRAW32 else self._next_64

    @property
    def _next_32(self) -> Callable[[Any], uint32]:
        """
        A callable with the same signature as as next_raw that always returns
        a random numpy 32-bit unsigned int.
        """

        def next_32(void_p: Any) -> uint32:
            return uint32(self.random_uint(32))

        return next_32

    @property
    def _next_64(self) -> Callable[[Any], uint64]:
        """
        A callable with the same signature as as next_raw that always returns
        a random numpy 64-bit unsigned int.
        """

        def next_64(void_p: Any) -> uint64:
            return uint64(self.random_uint(64))

        return next_64

    @property
    def _next_double(self) -> Callable[[Any], float64]:
        """
        A callable with the same signature as as next_raw that always return
        a random double in [0,1).
        """

        def next_double(void_p: Any) -> float64:
            return float64(self.random_double(1))

        return next_double



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import List, Union


###############################################################################
## REVERSE ENDIAN
###############################################################################
def reverse_endian(bitstrings: Union[str, List[str]]) -> Union[str, List[str]]:
    single_element: bool = type(bitstrings) is not list
    if single_element:
        bitstrings = [bitstrings]  # type: ignore
    reversed: List[str] = []
    for s in bitstrings:
        reversed.append(s[::-1])
    return reversed if not single_element else reversed[0]



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

###############################################################################
## IS BITSTRING
###############################################################################
def is_bitstring(bitstring: str) -> bool:
    """
    Returns `True` if the input str is a bitstring, `False` otherwise.

    PARAMETERS
    ----------
    bitstring: str
        The string to check.

    RETURNS
    -------
    out: bool
        `True` if input str is bitstring, `False` otherwise.

    RAISES
    ------
    TypeError
        If input bitstring is not str.
    """
    if not isinstance(bitstring, str):
        raise TypeError(f"Invalid bitstring type '{type(bitstring)}'")
    b = {"0", "1"}
    s = set(bitstring)
    return s.issubset(b)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

import struct
from typing import Any, Callable, Final, List, Optional, Tuple
from warnings import warn

from qiskit import BasicAer, QuantumCircuit, execute
from qiskit.providers import Backend, Job, Provider
from qiskit.providers.ibmq import IBMQError, least_busy
from qiskit.providers.models import BackendConfiguration
from qiskit.result import Counts, Result

from .caches import BasicCache as BitCache
from .quantum_bit_generator import QuantumBitGenerator

###############################################################################
## DEPRECATION WARNING -> warn(WARNING_MESSAGE, FutureWarning)
###############################################################################
WARNING_MESSAGE = "QiskitBitGenerator will be deprecated in version 1.0.0."
WARNING_MESSAGE += " Use QuantumBitGenerator and QiskitPlatform instead."


###############################################################################
## CUSTOM TYPES
###############################################################################
BackendFilter = Callable[[Backend], bool]


###############################################################################
## QISKIT BIT GENERATOR
###############################################################################
class QiskitBitGenerator(QuantumBitGenerator):
    """
    A quantum random bit-generator based on Qiskit, which can interface with
    NumPy's random module (e.g. to instantiate Generator objects). It
    implements an efficient strategy to retrieve random bits from IBMQ's quantum
    backends.

    PARAMETERS
    ----------
    provider: Optional[Provider] = None
        A Qiskit Provider object to access quantum backends.
        If `None` it defaults to BasicAer.
    backend: Optional[Backend] = None
        A Qiskit Backend object to produce random bits.
        If not `None`, `provider` will be ignored.
    backend_filter: Optional[BackendFilter] = None
        A Callable that takes in a Qiskit Backend object and returns `True`
        if it meets certain requirements, `False` otherwise. This is used
        to filter the list of available backends from which to dynamically
        choose on each request to the `provider` (if no `backend` is
        explicitly input). If `None` it defaults to
        `QiskitBitGenerator.default_backend_filter`.
    max_bits_per_request: int = 0
        A limit to the number of bits to be retrieved on each request to any
        Qiskit Backend. If less than one, no bound will be applied and the
        maximum allowed number of bits will be retrieved.
    ISRAW32: Final[bool] = False
        Toggle 32-bit BitGenerator mode. If `False` the BitGenerator will
        be 64-bit. This determines the number of bits returned by NumPy's
        `random_raw()` method, and the default number of bits to output on
        `random_uint()` and `random_double()`. Once an object is instantiated,
        this cannot be overridden.

    NOTES
    -----
    On each request to a backend, it retrieves as many bits as possible and
    stores them in a cache. This way, the number of internet connections
    leading to overheads is greatly reduced and, while the cache is loaded,
    random bits can be retrieved "instantaneously". The user can limit the
    number of bits to retrieve on each request through the
    `max_bits_per_request` parameter.
    Additionally, it always chooses the least busy backend from the list of
    machines available to the given provider. This list can be filtered by the
    user through the `backend_filter` parameter, which defaults to history-
    enabled non-simulators. If a Qiskit Backend is explicitly passed in as
    parameter, no backend selection will be performed: effectively ignoring any
    filters, or Qiskit Provider object passed.
    If neither `provider` nor `backend` are passed as inputs, it will default
    to running Qiskit BasicAer's 'qasm_simulator' locally.
    """

    _BACKEND_CONFIG_MASK: Final[dict] = {
        "backend_name": "",
        "credits_required": False,
        "local": True,
        "n_qubits": None,
        "simulator": True,
    }

    def __init__(
        self,
        provider: Optional[Provider] = None,
        backend: Optional[Backend] = None,
        backend_filter: Optional[BackendFilter] = None,
        max_bits_per_request: int = 0,
        ISRAW32: bool = False,
    ) -> None:
        warn(WARNING_MESSAGE, FutureWarning)
        if backend:
            provider = None
        elif provider:
            backend = self.get_best_backend(
                provider=provider,
                backend_filter=backend_filter,
            )
        else:
            backend = BasicAer.get_backend("qasm_simulator")
        self._provider: Optional[Provider] = provider
        self._backend: Backend = backend
        self._backend_filter: Optional[BackendFilter] = backend_filter
        self._set_mbpr(max_bits_per_request)
        self._ISRAW32: Final[bool] = ISRAW32  # type: ignore
        self._bitcache: BitCache = BitCache()
        super(QuantumBitGenerator, self).__init__(
            bits=self.BITS,
            next_raw=self._next_raw,
            next_32=self._next_32,
            next_64=self._next_64,
            next_double=self._next_double,
        )

    ########################## STATIC/CLASS METHODS ##########################
    @staticmethod
    def default_backend_filter(b: Backend) -> bool:
        """
        Default backend filter Callable.

        PARAMETERS
        ----------
        b: Backend
            A Qiskit Backend object.

        RETURNS
        -------
        out: bool
            `True` if `b` is not a simulator and has memory enabled, `False`
            otherwise.
        """
        config: BackendConfiguration = b.configuration()
        return config.memory and not config.simulator

    @classmethod
    def get_best_backend(
        cls,
        provider: Provider,
        backend_filter: Optional[BackendFilter] = None,
    ) -> Backend:
        """
        Returns the least busy backend available to an input provider, and
        according to certain filter(s).

        PARAMETERS
        ----------
        provider: Provider
            A Qiskit Provider object to access quantum backends.
        backend_filter: Optional[BackendFilter] = None
            A Callable that takes in a Qiskit Backend object and returns `True`
            if it meets certain requirements, `False` otherwise.
            If `None` it defaults to `cls.default_backend_filter`.

        RETURNS
        -------
        out: Backend
            Least busy backend from the filtered list of available backends.
        """
        if not backend_filter:
            backend_filter = cls.default_backend_filter
        backends: List[Backend] = provider.backends(filters=backend_filter)
        if not backends:
            raise IBMQError(
                "No backends matching the filtering critera on the \
                requested provider."
            )
        return least_busy(backends)

    ############################# PUBLIC METHODS #############################
    def flush_cache(self) -> None:
        """
        Erase the cache.

        RETURNS
        -------
        out: bool
            `True` if succeeds, `False` otherwise.
        """
        self._bitcache.flush()

    def load_cache(self, bitstring: str, flush: bool = False) -> None:
        """
        Load cache contents from bitstring.

        PARAMETERS
        ----------
        bitstring: str
            The bitstring to load to cache.
        flush: bool
            If `True` erase cache before loading.

        RETURNS
        -------
        out: bool
            `True` if succeeds, `False` otherwise.

        RAISES
        ------
        TypeError (push)
            If input bitstring is not str
        ValueError (push)
            If input bitstring is not a valid bitstring
        """
        if flush:
            self._bitcache.flush()
        self._bitcache.push(bitstring)

    def random_bitstring(self, n_bits: int = 0) -> str:
        """
        Returns a random bitstring of a given lenght. If less than one it
        defaults to the raw number of bits for the instance QiskitBitGenerator
        (i.e. 32 or 64).

        PARAMETERS
        ----------
        n_bits: int
            Number of bits to retrieve.

        RETURNS
        -------
        out: str
            Bitstring of lenght `n_bits`.
        """
        if n_bits < 1:
            n_bits = self.BITS
        while self._bitcache.size < n_bits:
            self._fetch_random_bits()
        return self._bitcache.pop(n_bits)

    def random_uint(self, n_bits: int = 0) -> int:
        """
        Returns a random unsigned int of a given size in bits.

        PARAMETERS
        ----------
        n_bits: int
            Number of bits to retrieve. If less than one it defaults to the raw
            number of bits for the instance QiskitBitGenerator (i.e. 32 or 64).

        RETURNS
        -------
        out: int
            Unsigned int of `n_bits` bits.
        """
        if n_bits < 1:
            n_bits = self.BITS
        return int(self.random_bitstring(n_bits), 2)

    def set_state(
        self,
        provider: Optional[Provider] = None,
        backend: Optional[Backend] = None,
        backend_filter: Optional[BackendFilter] = None,
        max_bits_per_request: Optional[int] = None,
    ) -> bool:
        """
        Override constructor parameters.
        Any change must be explicitly passed as input (i.e. not `None`).

        PARAMETERS
        ----------
        provider: Optional[Provider] = None
            Same as constructor.
        backend: Optional[Backend] = None
            Same as constructor.
        backend_filter: Optional[BackendFilter] = None
            Same as constructor.
        max_bits_per_request: Optional[int] = None
            Same as constructor.

        RETURNS
        -------
        out: bool
            `True` if any changes were made, `False` otherwise.
        """
        change: bool = False
        if max_bits_per_request is not None:
            change = True
            self._set_mbpr(max_bits_per_request)
        if backend_filter:
            change = True
            self._backend_filter = backend_filter
        if backend:
            change = True
            self._provider = None
            self._backend = backend
        elif provider:
            change = True
            self._provider = provider
            self._backend = self.get_best_backend(
                provider=provider,
                backend_filter=self._backend_filter,
            )
        return change

    ############################# PRIVATE METHODS #############################
    def _fetch_random_bits(self) -> bool:
        if self._provider:
            self._backend = self.get_best_backend(
                provider=self._provider,
                backend_filter=self._backend_filter,
            )
        circuits: List[QuantumCircuit] = [self._circuit] * self._experiments
        job: Job = execute(
            circuits,
            self._backend,
            shots=self._shots,
            memory=self._memory,
        )
        result: Result = job.result()
        bitstring: str = self._parse_result(result)
        self._bitcache.push(bitstring)
        return True

    def _parse_backend_config(self, backend_config: dict) -> dict:
        keys = backend_config.keys()
        config: dict = {}
        for k, v in self._BACKEND_CONFIG_MASK.items():
            config[k] = backend_config[k] if k in keys else v
        return config

    def _parse_result(self, result: Result) -> str:
        measurements: List[str] = []
        if self._memory:
            for e in range(self._experiments):
                measurements += result.get_memory(e)
        else:
            cts = result.get_counts()
            counts: List[Counts] = [cts] if type(cts) != list else cts
            for c in counts:
                measurements += [k for k, v in c.items() if v == 1]
        bitstring: str = ""
        for m in measurements:
            bitstring += m
        return bitstring

    def _set_mbpr(self, max_bits_per_request: int) -> bool:
        self._max_bits_per_request = (
            max_bits_per_request if max_bits_per_request > 0 else 0
        )
        return True

    ############################ PUBLIC PROPERTIES ############################
    @property
    def state(self) -> dict:
        """
        Parsed information about the current state of the QiskitBitGenerator.
        """
        s: dict = {
            "BITS": self.BITS,
            "job_config": self._job_config,
            "backend_config": self._parse_backend_config(self._backend_config),
            "dynamic_backend": {
                "filter": "Custom" if self._backend_filter else "Default",
            },
            "bitcache": {"size": self._bitcache.size},
        }
        if not self._provider:
            s.pop("dynamic_backend")
        return s

    @state.setter
    def state(self, value: dict) -> None:
        """
        An alternative form of `set_state()` parsed from dict equality.
        """
        self.set_state(**value)

    ########################### PRIVATE PROPERTIES ###########################
    @property
    def _backend_config(self) -> dict:
        return self._backend.configuration().to_dict()

    @property
    def _circuit(self) -> QuantumCircuit:
        n_qubits: int = self._n_qubits
        circuit: QuantumCircuit = QuantumCircuit(n_qubits)
        circuit.h(range(n_qubits))
        circuit.measure_all()
        return circuit

    @property
    def _experiments(self) -> int:
        n_qubits, shots, experiments = self._job_partition
        return experiments

    @property
    def _job_config(self) -> dict:
        return {
            "max_bits_per_request": self._max_bits_per_request or None,
            "bits_per_request": self._n_qubits
            * self._shots
            * self._experiments,
            "n_qubits": self._n_qubits,
            "shots": self._shots,
            "experiments": self._experiments,
        }

    @property
    def _job_partition(self) -> Tuple[int, int, int]:
        backend_config: dict = self._backend_config
        experiments: int = (
            backend_config["max_experiments"]
            if backend_config.__contains__("max_experiments")
            and backend_config["max_experiments"]
            else 1
        )
        shots: int = (
            backend_config["max_shots"]
            if backend_config.__contains__("max_shots")
            and backend_config["max_shots"]
            and backend_config.__contains__("memory")
            and backend_config["memory"]
            else 1
        )
        n_qubits: int = (
            backend_config["n_qubits"]
            if backend_config.__contains__("n_qubits")
            and backend_config["n_qubits"]
            else 1
        )
        max_bits_per_request: int = self._max_bits_per_request
        if max_bits_per_request > n_qubits:
            experiments = min(
                experiments,
                max_bits_per_request // (shots * n_qubits) + 1,
            )
            shots = min(
                shots,
                max_bits_per_request // (experiments * n_qubits),
            )
        elif max_bits_per_request > 0:
            experiments = 1
            shots = 1
            n_qubits = max_bits_per_request
        return n_qubits, shots, experiments

    @property
    def _memory(self) -> bool:
        return True if self._shots > 1 else False

    @property
    def _n_qubits(self) -> int:
        n_qubits, shots, experiments = self._job_partition
        return n_qubits

    @property
    def _shots(self) -> int:
        n_qubits, shots, experiments = self._job_partition
        return shots



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import List, Optional
from warnings import warn

from qiskit import QuantumCircuit as QiskitQuantumCircuit
from qiskit import execute
from qiskit.providers import Job
from qiskit.result import Counts, Result

from ...helpers import compute_bounded_factorization, reverse_endian
from ..job import QuantumJob
from .backend import QiskitBackend
from .circuit import QiskitCircuit


###############################################################################
## QISKIT JOB
###############################################################################
class QiskitJob(QuantumJob):
    def __init__(
        self,
        circuit: QiskitCircuit,
        backend: QiskitBackend,
        num_measurements: Optional[int] = None,
    ) -> None:
        self.backend: QiskitBackend = backend
        self.circuit: QiskitCircuit = circuit
        self.num_measurements: int = num_measurements  # type: ignore
        self._base_job: Optional[Job] = None

    ############################### PUBLIC API ###############################
    @property
    def backend(self) -> QiskitBackend:
        return self._backend

    @backend.setter
    def backend(self, backend: QiskitBackend) -> None:
        self._backend: QiskitBackend = backend

    @property
    def circuit(self) -> QiskitCircuit:
        return self._circuit

    @circuit.setter
    def circuit(self, circuit: QiskitCircuit) -> None:
        if self.backend.max_qubits < circuit.num_qubits:
            raise RuntimeError(
                f"Failed to assign QiskitCircuit for QiskitJob. Number of \
                qubits in QiskitCircuit unsupported by this job's Backend: \
                {self.backend.max_qubits}<{circuit.num_qubits}."
            )
        self._circuit: QiskitCircuit = circuit

    @property
    def num_measurements(self) -> int:
        return self._shots * self._experiments

    @num_measurements.setter
    def num_measurements(self, num_measurements: Optional[int]) -> None:
        num_measurements = (
            num_measurements
            if num_measurements
            and type(num_measurements) is int
            and 0 < num_measurements
            else self.backend.max_measurements
        )
        if self.backend.max_measurements < num_measurements:
            warn(
                f"Number of measurements unsupported by the job's Backend: \
                {self.backend.max_measurements}<{num_measurements}. \
                Using max_measurements instead.",
                UserWarning,
            )
            num_measurements = self.backend.max_measurements
        self._shots, self._experiments = compute_bounded_factorization(
            num_measurements,
            self.backend.max_shots,
            self.backend.max_experiments,
        )

    def execute(self) -> List[str]:
        circuits: List[QiskitQuantumCircuit] = [
            self.circuit
        ] * self._experiments
        self._base_job = execute(
            circuits,
            self.backend,
            shots=self._shots,
            memory=self._requires_memory,
        )
        result: Result = self._base_job.result()
        return self._parse_result(result)

    ############################### PRIVATE API ###############################
    @property
    def _experiments(self) -> int:
        return self.__experiments

    @_experiments.setter
    def _experiments(self, experiments: Optional[int]) -> None:
        self.__experiments: int = (
            experiments
            if type(experiments) is int
            and 0 < experiments < self.backend.max_experiments  # type: ignore
            else self.backend.max_experiments
        )

    @property
    def _requires_memory(self) -> bool:
        return True if self._shots > 1 else False

    @property
    def _shots(self) -> int:
        return self.__shots

    @_shots.setter
    def _shots(self, shots: Optional[int]) -> None:
        self.__shots: int = (
            shots
            if type(shots) is int
            and 0 < shots < self.backend.max_shots  # type: ignore
            else self.backend.max_shots
        )

    def _parse_result(self, result: Result) -> List[str]:
        measurements: List[str] = []
        if self._requires_memory:
            for e in range(self._experiments):
                measurements += result.get_memory(e)
        else:
            cts = result.get_counts()
            counts: List[Counts] = [cts] if type(cts) != list else cts
            for c in counts:
                measurements += [k for k, v in c.items() if v == 1]
        return reverse_endian(measurements)  # type: ignore



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import Callable, List, Optional, Tuple

from qiskit import BasicAer
from qiskit.providers import BackendV1 as Backend
from qiskit.providers import Provider
from qiskit.providers.ibmq import IBMQError, least_busy
from qiskit.providers.models import BackendConfiguration

from ...protocols import ProtocolResult, QuantumProtocol
from ..platform import QuantumPlatform
from .backend import QiskitBackend
from .circuit import QiskitCircuit
from .job import QiskitJob

###############################################################################
## CUSTOM TYPES
###############################################################################
BackendFilter = Callable[[Backend], bool]


###############################################################################
## QISKIT PLATFORM
###############################################################################
class QiskitPlatform(QuantumPlatform):
    def __init__(
        self,
        provider: Optional[Provider] = None,
        backend: Optional[Backend] = None,
        backend_filter: Optional[BackendFilter] = None,
    ) -> None:
        if backend:
            provider = None
        elif provider:
            backend = self._get_best_backend(
                provider=provider,
                backend_filter=backend_filter,
            )
        else:
            backend = BasicAer.get_backend("qasm_simulator")
        self.provider: Optional[Provider] = provider
        self.backend: Backend = backend
        self.backend_filter: BackendFilter = backend_filter  # type: ignore

    ############################### PUBLIC API ###############################
    @property
    def backend(self) -> QiskitBackend:
        return self._backend

    @backend.setter
    def backend(self, backend: Backend) -> None:
        self._backend: Backend = backend

    @property
    def backend_filter(self) -> BackendFilter:
        return self._backend_filter

    @backend_filter.setter
    def backend_filter(self, backend_filter: Optional[BackendFilter]) -> None:
        self._backend_filter = backend_filter or self.default_backend_filter

    @property
    def provider(self) -> Optional[Provider]:
        return self._provider

    @provider.setter
    def provider(self, provider: Optional[Provider]) -> None:
        self._provider: Provider = provider

    @staticmethod
    def default_backend_filter(b: Backend) -> bool:
        config: BackendConfiguration = b.configuration()
        return config.memory and not config.simulator

    def create_circuit(self, num_qubits: int) -> QiskitCircuit:  # type: ignore
        return QiskitCircuit(num_qubits)

    def create_job(  # type: ignore
        self,
        circuit: QiskitCircuit,
        backend: QiskitBackend,
        num_measurements: Optional[int],
    ) -> QiskitJob:
        return QiskitJob(circuit, backend, num_measurements)

    def fetch_random_bits(
        self, protocol: QuantumProtocol, max_bits: Optional[int] = None
    ) -> str:
        max_bits = max_bits if max_bits and max_bits > 0 else None
        result: ProtocolResult = protocol.run(self, max_bits)
        return result.bitstring

    def retrieve_backend(
        self, max_bits_per_request: Optional[int] = None
    ) -> QiskitBackend:
        if self.provider:
            self.backend = self._get_best_backend(
                provider=self.provider,
                backend_filter=self.backend_filter,
            )
        return QiskitBackend(self.backend, max_bits_per_request)

    ############################### PRIVATE API ###############################
    @classmethod
    def _get_best_backend(
        cls,
        provider: Provider,
        backend_filter: Optional[BackendFilter] = None,
    ) -> Backend:
        if not backend_filter:
            backend_filter = cls.default_backend_filter
        backends: List[Backend] = provider.backends(filters=backend_filter)
        if not backends:
            raise IBMQError(  # TODO
                "No backends matching the filtering critera on the \
                requested provider."
            )
        return least_busy(backends)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 14, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

import math
from base64 import b32encode, b64encode
from struct import pack, unpack
from typing import Optional
from warnings import warn

from .quantum_bit_generator import QuantumBitGenerator

DEPRECATION_WARNING = " will be deprecated in version 1.0.0."


###############################################################################
## QRNG
###############################################################################
class Qrng:
    """
    An integrated implementation of the QRNG PyPI package which makes use of
    QRAND's enhanced random bit generation and caching capabilities.

    PARAMETERS
    ----------
    quantum_bit_generator: QuantumBitGenerator
        A QuantumBitGenerator instance object to handle random number production.

    ATTRIBUTES
    ----------
    state: dict
        The state of the Qrng object.

    COPYRIGHT NOTICE
    ----------------
    Source: https://github.com/ozanerhansha/qRNG
    License: GNU GENERAL PUBLIC LICENSE VERSION 3
    Changes:
        - Delete IBMQ log-in logic
        - Replace random bit generation and caching logic
        - Add default parameter values
        - Rename internal variables
        - Add static type hints
    """

    def __init__(self, quantum_bit_generator: QuantumBitGenerator):
        self._quantum_bit_generator = quantum_bit_generator

    ############################# PUBLIC METHODS #############################
    def get_bit_string(self, num_bits: int = 0):
        """
        Returns a random bitstring of a given lenght. If less than one it
        defaults to the raw number of bits for its internal
        quantum_bit_generator (i.e. 32 or 64).

        PARAMETERS
        ----------
        num_bits: int, default 0
            Number of bits to retrieve.

        RETURNS
        -------
        out: str
            Bitstring of lenght `num_bits`.
        """
        ## DEPRECATION WARNING
        WARNING_MESSAGE = "get_bit_string()" + DEPRECATION_WARNING
        warn(WARNING_MESSAGE, FutureWarning)
        return self._quantum_bit_generator.random_bitstring(num_bits)

    def get_random_base32(self, num_bits: int = 0) -> str:
        """
        Returns a random base32 encoded string from a num_bits uniform
        distribution. If less than one it defaults to the raw number of
        bits for its internal quantum_bit_generator (i.e. 32 or 64).

        RETURNS
        -------
        out: str
            Random base32 enocoded string.
        """
        b: bytes = self.get_random_bytes(num_bits)
        enc: bytes = b32encode(b)
        return enc.decode("utf-8")

    def get_random_base64(self, num_bits: int = 0) -> str:
        """
        Returns a random base64 encoded string from a num_bits uniform
        distribution. If less than one it defaults to the raw number of
        bits for its internal quantum_bit_generator (i.e. 32 or 64).

        RETURNS
        -------
        out: str
            Random base64 enocoded string.
        """
        b: bytes = self.get_random_bytes(num_bits)
        enc: bytes = b64encode(b)
        return enc.decode("utf-8")

    def get_random_bitstring(self, num_bits: int = 0):
        """
        Returns a random bitstring of a given lenght. If less than one it
        defaults to the raw number of bits for its internal
        quantum_bit_generator (i.e. 32 or 64).

        PARAMETERS
        ----------
        num_bits: int, default 0
            Number of bits to retrieve.

        RETURNS
        -------
        out: str
            Bitstring of lenght `num_bits`.
        """
        return self._quantum_bit_generator.random_bitstring(num_bits)

    def get_random_bytes(self, num_bits: int = 0) -> bytes:
        """
        Returns a bytes object from a num_bits uniform distribution.

        RETURNS
        -------
        out: bytes
            Random bytes object of length (num_bits + 7) // 8.
        """
        if num_bits < 1:
            num_bits = self._quantum_bit_generator.BITS
        uint: int = self._quantum_bit_generator.random_uint(num_bits)
        num_bytes: int = (num_bits + 7) // 8
        return uint.to_bytes(num_bytes, "big")

    def get_random_complex_polar(
        self, r: float = 1, theta: float = 2 * math.pi
    ):
        """
        Returns a random complex in rectangular form from a given polar range.
        If no max radius give, 1 is used. If no max angle given, 2pi used.

        PARAMETERS
        ----------
        r: float, default 1
            Real lower bound for the random number.
        theta: float, default 2pi
            Real strict upper bound for the random number.

        RETURNS
        -------
        out: complex
            Random complex in the range [0,r) * exp{ j[0,theta) }.
        """
        r0: float = r * math.sqrt(self.get_random_double(0, 1))
        theta0: float = self.get_random_double(0, theta)
        return r0 * math.cos(theta0) + r0 * math.sin(theta0) * 1j

    def get_random_complex_rect(
        self,
        r1: float = -1,
        r2: float = +1,
        i1: Optional[float] = None,
        i2: Optional[float] = None,
    ):
        """
        Returns a random complex with both real and imaginary parts from the
        given ranges. Default real range [-1,1). If no imaginary range
        specified, real range used.

        PARAMETERS
        ----------
        r1: float, default -1
            Real lower bound for the random number.
        r2: float, default +1
            Real strict upper bound for the random number.
        i1: float, default None
            Imaginary lower bound for the random number.
        i2: float, default None
            Imaginary strict upper bound for the random number.

        RETURNS
        -------
        out: complex
            Random complex in the range [r1,r2) + j[i1,i2).
        """
        re: float = self.get_random_double(r1, r2)
        if i1 is None or i2 is None:
            im: float = self.get_random_double(r1, r2)
        else:
            im = self.get_random_double(i1, i2)
        return re + im * 1j

    def get_random_decimal(self, num_bits: int = 0) -> str:
        """
        Returns a random decimal base encoded string from a num_bits uniform
        distribution. If less than one it defaults to the raw number of
        bits for its internal quantum_bit_generator (i.e. 32 or 64).

        RETURNS
        -------
        out: str
            Random decimal base enocoded string.
        """
        uint: int = self._quantum_bit_generator.random_uint(num_bits)
        return f"{uint:d}"

    def get_random_double(self, min: float = -1, max: float = +1):
        """
        Returns a random double from a uniform distribution in the range
        [min,max). Default range [-1,1).

        PARAMETERS
        ----------
        min: float, default -1
            Lower bound for the random number.
        max: float, default +1
            Strict upper bound for the random number.

        RETURNS
        -------
        out: float
            Random float in the range [min,max).
        """
        delta: float = max - min
        shifted: float = self._quantum_bit_generator.random_double(delta)
        return shifted + min

    def get_random_float(self, min: float = -1, max: float = +1):
        """
        Returns a random float from a uniform distribution in the range
        [min,max). Default range [-1,1).

        PARAMETERS
        ----------
        min: float, default -1
            Lower bound for the random number.
        max: float, default +1
            Strict upper bound for the random number.

        RETURNS
        -------
        out: float
            Random float in the range [min,max).
        """
        unpacked = 0x3F800000 | self.get_random_int32() >> 9
        packed = pack("I", unpacked)
        value = unpack("f", packed)[0] - 1.0
        return (max - min) * value + min

    def get_random_hex(self, num_bits: int = 0) -> str:
        """
        Returns a random hex base encoded string from a num_bits uniform
        distribution. If less than one it defaults to the raw number of
        bits for its internal quantum_bit_generator (i.e. 32 or 64).

        RETURNS
        -------
        out: str
            Random hex enocoded string.
        """
        uint: int = self._quantum_bit_generator.random_uint(num_bits)
        return f"{uint:X}"

    def get_random_int(self, min: int = -1, max: int = +1):
        """
        Returns a random integer between and including [min, max]. Default
        range [-1,1].

        PARAMETERS
        ----------
        min: int, default -1
            Lower bound for the random int.
        max: int, default +1
            Upper bound for the random int.

        RETURNS
        -------
        out: int
            Random int in the range [min,max].
        """
        delta: int = max - min
        num_bits: int = math.floor(math.log(delta, 2)) + 1
        shifted: int = self._quantum_bit_generator.random_uint(num_bits)
        while shifted > delta:
            shifted = self._quantum_bit_generator.random_uint(num_bits)
        return shifted + min

    def get_random_int32(self):
        """
        Returns a random 32 bit unsigned integer from a uniform distribution.

        RETURNS
        -------
        out: int
            Random 32 bit unsigned int.
        """
        return self._quantum_bit_generator.random_uint(32)

    def get_random_int64(self):
        """
        Returns a random 32 bit unsigned integer from a uniform distribution.

        RETURNS
        -------
        out: int
            Random 64 bit unsigned int.
        """
        return self._quantum_bit_generator.random_uint(64)

    def get_random_octal(self, num_bits: int = 0) -> str:
        """
        Returns a random octal base encoded string from a num_bits uniform
        distribution. If less than one it defaults to the raw number of
        bits for its internal quantum_bit_generator (i.e. 32 or 64).

        RETURNS
        -------
        out: str
            Random octal base enocoded string.
        """
        uint: int = self._quantum_bit_generator.random_uint(num_bits)
        return f"{uint:o}"

    def get_random_uint(self, num_bits: int = 0) -> int:
        """
        Returns a random unsigned int of a given size in bits.

        PARAMETERS
        ----------
        num_bits: int
            Number of bits to retrieve. If less than one it defaults to the raw
            number of bits for the instance QiskitBitGenerator (i.e. 32 or 64).

        RETURNS
        -------
        out: int
            Unsigned int of `num_bits` bits.
        """
        return self._quantum_bit_generator.random_uint(num_bits)

    ############################ PUBLIC PROPERTIES ############################
    @property
    def state(self) -> dict:
        """
        The state of the Qrng object.
        """
        ## DEPRECATION WARNING
        WARNING_MESSAGE = "state()" + DEPRECATION_WARNING
        warn(WARNING_MESSAGE, FutureWarning)
        return {
            "quantum_bit_generator": self._quantum_bit_generator.state,
        }



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod


###############################################################################
## QUANTUM CIRCUIT INTERFACE (ADAPTER)
###############################################################################
class QuantumCircuit(ABC):
    @abstractmethod
    def __init__(self, num_qubits: int) -> None:
        pass

    @property
    @abstractmethod
    def num_qubits(self) -> int:
        pass

    ############################## SPECIAL GATES ##############################
    @abstractmethod
    def measure(self, target_qubit: int) -> None:
        pass

    ########################### SINGLE QUBIT GATES ###########################
    @abstractmethod
    def h(self, target_qubit: int) -> None:
        pass

    @abstractmethod
    def rx(self, radians: float, target_qubit: int) -> None:
        pass

    @abstractmethod
    def ry(self, radians: float, target_qubit: int) -> None:
        pass

    @abstractmethod
    def rz(self, radians: float, target_qubit: int) -> None:
        pass

    @abstractmethod
    def s(self, target_qubit: int) -> None:
        pass

    @abstractmethod
    def t(self, target_qubit: int) -> None:
        pass

    @abstractmethod
    def u1(self, theta: float, target_qubit: int) -> None:
        pass

    @abstractmethod
    def u2(self, phi: float, lam: float, target_qubit: int) -> None:
        pass

    @abstractmethod
    def u3(
        self, theta: float, phi: float, lam: float, target_qubit: int
    ) -> None:
        pass

    @abstractmethod
    def x(self, target_qubit: int) -> None:
        pass

    @abstractmethod
    def y(self, target_qubit: int) -> None:
        pass

    @abstractmethod
    def z(self, target_qubit: int) -> None:
        pass

    ############################# TWO QUBIT GATES #############################
    @abstractmethod
    def cs(self, control_qubit: int, target_qubit: int) -> None:
        pass

    @abstractmethod
    def cx(self, control_qubit: int, target_qubit: int) -> None:
        pass

    @abstractmethod
    def cz(self, control_qubit: int, target_qubit: int) -> None:
        pass

    @abstractmethod
    def swap(self, target_qubit_1: int, target_qubit_2: int) -> None:
        pass

    ############################ THREE QUBIT GATES ############################
    @abstractmethod
    def ccx(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        pass

    @abstractmethod
    def cswap(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        pass

    ############################## ALIAS GATES ##############################
    def cnot(self, control_qubit: int, target_qubit: int) -> None:
        self.cx(control_qubit, target_qubit)

    def cphase(self, control_qubit: int, target_qubit: int) -> None:
        self.cs(control_qubit, target_qubit)

    def fredkin(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        self.cswap(control_qubit_1, control_qubit_2, target_qubit)

    def hadamard(self, target_qubit: int) -> None:
        self.h(target_qubit)

    def phase(self, target_qubit: int) -> None:
        self.s(target_qubit)

    def pi8(self, target_qubit: int) -> None:
        self.t(target_qubit)

    def toffoli(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        self.ccx(control_qubit_1, control_qubit_2, target_qubit)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import List, Literal, Optional

from ..platforms.backend import QuantumBackend
from ..platforms.circuit import QuantumCircuit
from ..platforms.factory import QuantumFactory
from ..platforms.job import QuantumJob
from .protocol import BareQuantumProtocol
from .result import BasicResult, ProtocolResult


###############################################################################
## HADAMARD PROTOCOL
###############################################################################
class HadamardProtocol(BareQuantumProtocol):
    ############################### PUBLIC API ###############################
    def run(
        self, factory: QuantumFactory, max_bits: Optional[int] = None
    ) -> ProtocolResult:
        backend: QuantumBackend = factory.retrieve_backend(max_bits)
        num_qubits, num_measurements = backend.job_partition
        circuit: QuantumCircuit = factory.create_circuit(num_qubits)
        circuit = self._assemble_quantum_circuit(circuit)
        job: QuantumJob = factory.create_job(
            circuit, backend, num_measurements
        )
        output: List[str] = job.execute()
        return self._parse_output(output)

    def verify(self) -> Literal[False]:
        return False

    ############################### PRIVATE API ###############################
    def _assemble_quantum_circuit(
        self, circuit: QuantumCircuit
    ) -> QuantumCircuit:
        for q in range(circuit.num_qubits):
            circuit.h(q)
            circuit.measure(q)
        return circuit

    def _parse_output(self, output: List[str]) -> ProtocolResult:
        bitstring: str = ""
        for measurement in output:
            bitstring += measurement
        return BasicResult(bitstring)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from ..circuit import QuantumCircuit


###############################################################################
## QSHARP CIRCUIT
###############################################################################
class QsharpCircuit(QuantumCircuit):
    def __init__(self, num_qubits: int) -> None:
        self.ERROR_MSG = f"{self.__class__.__name__}"  # TODO
        raise NotImplementedError(self.ERROR_MSG)

    @property
    def num_qubits(self) -> int:
        raise NotImplementedError(self.ERROR_MSG)

    ############################## SPECIAL GATES ##############################
    def measure(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    ########################### SINGLE QUBIT GATES ###########################
    def h(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def rx(self, radians: float, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def ry(self, radians: float, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def rz(self, radians: float, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def s(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def t(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def u1(self, theta: float, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def u2(self, phi: float, lam: float, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def u3(
        self, theta: float, phi: float, lam: float, target_qubit: int
    ) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def x(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def y(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def z(self, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    ############################# TWO QUBIT GATES #############################
    def cs(self, control_qubit: int, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def cx(self, control_qubit: int, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def cz(self, control_qubit: int, target_qubit: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def swap(self, target_qubit_1: int, target_qubit_2: int) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    ############################ THREE QUBIT GATES ############################
    def ccx(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        raise NotImplementedError(self.ERROR_MSG)

    def cswap(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        raise NotImplementedError(self.ERROR_MSG)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod

from ..helpers import is_bitstring


###############################################################################
## BIT CACHE INTERFACE
###############################################################################
class BitCache(ABC):
    """
    BitCache first in, first out (FIFO) data structure.
    """

    ############################### PUBLIC API ###############################
    @property
    @abstractmethod
    def size(self) -> int:
        """
        The number of bits currently stored in the BitCache.
        """
        pass

    @abstractmethod
    def dump(self) -> str:
        """
        Outputs all the contents in the cache without erasing.

        RETURNS
        -------
        out: str
            Cache contents.
        """
        pass

    @abstractmethod
    def flush(self) -> None:
        """
        Erases the cache.
        """
        pass

    @abstractmethod
    def pop(self, n: int) -> str:
        """
        Returns a size `n` bitstring removing it from the top of the cache.

        PARAMETERS
        ----------
        n: int
            Number of bits to retrieve

        RETURNS
        -------
        out: str
            Size `n` bitstring.

        RAISES
        ------
        ValueError
            If input is greater than cache size, or less than one.
        """
        pass

    @abstractmethod
    def push(self, bitstring: str) -> None:
        """
        Inserts bitstring at the end of the cache.

        PARAMETERS
        ----------
        bitstring: str
            The bitstring to insert.

        RAISES
        ------
        TypeError (is_bitstring)
            If input bitstring is not str
        ValueError
            If input bitstring is not a valid bitstring
        """
        pass



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 13, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional


###############################################################################
## PROTOCOL RESULT INTERFACE
###############################################################################
class ProtocolResult(ABC):
    @property
    @abstractmethod
    def bitstring(self) -> str:
        pass

    @property
    @abstractmethod
    def validation_token(self) -> str:
        pass

    @abstractmethod
    def erase(self) -> None:
        pass


###############################################################################
## SIMPLE RESULT
###############################################################################
class BasicResult(ProtocolResult):
    def __init__(
        self, bitstring: str, validation_token: Optional[str] = None
    ) -> None:
        self._bitstring: str = bitstring
        self._validation_token: str = validation_token or bitstring

    ############################### PUBLIC API ###############################
    @property
    def bitstring(self) -> str:
        return self._bitstring

    @property
    def validation_token(self) -> str:
        return self._validation_token

    def erase(self) -> None:
        self._bitstring = ""
        self._validation_token = ""



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import Tuple


###############################################################################
## COMPUTE BOUNDED FACTORIZATION
###############################################################################
def compute_bounded_factorization(
    n: int, bound_A: int, bound_B: int
) -> Tuple[int, int]:
    if bound_A * bound_B < n:
        return bound_A, bound_B
    swapped: bool = bound_A > bound_B
    bound_A, bound_B = sorted([bound_A, bound_B])
    final_b: int = bound_B
    final_a: int = n // final_b
    final_delta: int = n - final_a * final_b
    a: int = final_a + 1
    b: int = n // a
    delta: int = n - a * b
    while a <= bound_A and a <= b and final_delta != 0:
        if delta < final_delta:
            final_a, final_b, final_delta = a, b, delta
        a += 1
        b = n // a
        delta = n - a * b
    return (final_b, final_a) if swapped else (final_a, final_b)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import Callable, Optional

from qiskit.providers import BackendV1 as Backend
from qiskit.providers import Job, Options

from ..backend import QuantumBackend


###############################################################################
## QISKIT BACKEND (DECORATOR)
###############################################################################
class QiskitBackend(QuantumBackend, Backend):
    def __init__(
        self, backend: Backend, max_bits_per_request: Optional[int] = None
    ) -> None:
        self._base_backend: Backend = backend
        super(QuantumBackend, self).__init__(
            backend.configuration(), backend.provider()
        )
        self._options: Options = backend._options
        try:
            mbpr = backend.max_bits_per_request
        except AttributeError:
            mbpr = None
        self.max_bits_per_request = max_bits_per_request or mbpr

    ############################### PUBLIC API ###############################
    @property
    def configuration_dict(self) -> dict:
        return self.configuration().to_dict()

    @property
    def max_experiments(self) -> int:
        return (
            self.configuration_dict["max_experiments"]
            if self.configuration_dict.__contains__("max_experiments")
            and self.configuration_dict["max_experiments"]
            else 1
        )

    @property
    def max_measurements(self) -> int:
        return self.max_shots * self.max_experiments

    @property
    def max_qubits(self) -> int:
        return (
            self.configuration_dict["num_qubits"]
            if self.configuration_dict.__contains__("num_qubits")
            and self.configuration_dict["num_qubits"]
            else 1
        )

    @property
    def max_shots(self) -> int:
        return (
            self.configuration_dict["max_shots"]
            if self.configuration_dict.__contains__("max_shots")
            and self.configuration_dict["max_shots"]
            and self.memory
            else 1
        )

    @property
    def memory(self) -> bool:
        return (
            self.configuration_dict["memory"]
            if self.configuration_dict.__contains__("memory")
            else False
        )

    ############################ QISKIT INTERFACE ############################
    def _default_options(self) -> Options:
        return self._base_backend._default_options()

    def run(self, run_input, **options) -> Job:
        return self._base_backend.run(run_input, **options)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: April 20, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from .circuit import QuantumCircuit
from .cirq import CirqCircuit, CirqJob, CirqPlatform
from .job import QuantumJob
from .platform import QuantumPlatform
from .qiskit import QiskitCircuit, QiskitJob, QiskitPlatform
from .qsharp import QsharpCircuit, QsharpJob, QsharpPlatform

__all__ = [
    "QuantumCircuit",
    "QuantumJob",
    "QuantumPlatform",
    "CirqCircuit",
    "CirqJob",
    "CirqPlatform",
    "QiskitCircuit",
    "QiskitJob",
    "QiskitPlatform",
    "QsharpCircuit",
    "QsharpJob",
    "QsharpPlatform",
]



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod
from typing import Literal, Optional

from ..platforms.factory import QuantumFactory
from ..validation import ValidationStrategy
from .result import ProtocolResult


###############################################################################
## QUANTUM PROTOCOL INTERFACE (STRATEGY AND DECORATOR)
###############################################################################
class QuantumProtocol(ABC):
    ############################ STRATEGY PATTERN ############################
    @abstractmethod
    def run(
        self, factory: QuantumFactory, max_bits: Optional[int] = None
    ) -> ProtocolResult:
        pass

    @abstractmethod
    def verify(self) -> bool:
        pass

    ############################ DECORATOR PATTERN ############################
    @property
    @abstractmethod
    def base_protocol(self) -> Optional["QuantumProtocol"]:
        pass

    @abstractmethod
    def validate(self, result: ProtocolResult) -> bool:
        pass


###############################################################################
## BASE QUANTUM PROTOCOL INTERFACE
###############################################################################
class BareQuantumProtocol(QuantumProtocol):
    ############################ STRATEGY PATTERN ############################
    @abstractmethod
    def run(
        self, factory: QuantumFactory, max_bits: Optional[int] = None
    ) -> ProtocolResult:
        pass

    @abstractmethod
    def verify(self) -> bool:
        pass

    ############################ DECORATOR PATTERN ############################
    @property
    def base_protocol(self) -> Literal[None]:
        return None

    def validate(self, result: ProtocolResult) -> Literal[False]:
        return False


###############################################################################
## VALIDATION DECORATOR
###############################################################################
class ValidationDecorator(QuantumProtocol):
    def __init__(
        self,
        base_protocol: QuantumProtocol,
        validation_strategy: ValidationStrategy,
    ) -> None:
        self.base_protocol: QuantumProtocol = base_protocol
        self.validation_strategy: ValidationStrategy = validation_strategy

    ############################ STRATEGY PATTERN ############################
    def run(
        self, factory: QuantumFactory, max_bits: Optional[int] = None
    ) -> ProtocolResult:
        result: ProtocolResult = self.base_protocol.run(factory, max_bits)
        if not self._validate_layer(result):
            result.erase()
        return result

    def verify(self) -> bool:
        return self.base_protocol.verify()

    ############################ DECORATOR PATTERN ############################
    @property
    def base_protocol(self) -> QuantumProtocol:
        return self._base_protocol

    @base_protocol.setter
    def base_protocol(self, protocol: QuantumProtocol) -> None:
        self._base_protocol: QuantumProtocol = protocol

    def validate(self, result: ProtocolResult) -> bool:
        valid: bool = self._validate_layer(result)
        if self.base_protocol.base_protocol is None:
            return valid
        return valid and self.base_protocol.validate(result)

    ############################### PUBLIC API ###############################
    @property
    def validation_strategy(self) -> ValidationStrategy:
        return self._validation_strategy

    @validation_strategy.setter
    def validation_strategy(self, strategy: ValidationStrategy) -> None:
        self._validation_strategy: ValidationStrategy = strategy

    ############################### PRIVATE API ###############################
    def _validate_layer(self, result: ProtocolResult) -> bool:
        validation_token: str = result.validation_token
        return self._validation_strategy.validate(validation_token)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional, Tuple

from ..helpers import compute_bounded_factorization


###############################################################################
## QUANTUM BACKEND BASE INTERFACE (ADAPTER)
###############################################################################
class QuantumBackend(ABC):
    ################################ ABSTRACT ################################
    @property
    @abstractmethod
    def max_measurements(self) -> int:
        pass

    @property
    @abstractmethod
    def max_qubits(self) -> int:
        pass

    ################################ CONCRETE ################################
    @property
    def job_partition(self) -> Tuple[int, int]:
        return compute_bounded_factorization(
            self.max_bits_per_request, self.max_qubits, self.max_measurements
        )

    @property
    def max_bits_per_request(self) -> int:
        try:
            mbpr = self._max_bits_per_request
        except AttributeError:
            mbpr = None
        max_allowed: int = self.max_qubits * self.max_measurements
        return mbpr if mbpr and 0 < mbpr < max_allowed else max_allowed

    @max_bits_per_request.setter
    def max_bits_per_request(self, mbpr: Optional[int]) -> None:
        mbpr = mbpr or 0
        if not isinstance(mbpr, int):
            raise TypeError(f"Expected int instance, {type(mbpr)} found")
        self._max_bits_per_request: Optional[int] = mbpr if mbpr > 0 else None



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from .backend import QsharpBackend
from .circuit import QsharpCircuit
from .job import QsharpJob
from .platform import QsharpPlatform

__all__ = [
    "QsharpBackend",
    "QsharpCircuit",
    "QsharpJob",
    "QsharpPlatform",
]



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Optional

from .backend import QuantumBackend
from .circuit import QuantumCircuit


###############################################################################
## QUANTUM JOB INTERFACE (ADAPTER)
###############################################################################
class QuantumJob(ABC):
    @property
    @abstractmethod
    def backend(self) -> QuantumBackend:
        pass

    @backend.setter
    def backend(self, backend: QuantumBackend) -> None:
        pass

    @property
    @abstractmethod
    def circuit(self) -> QuantumCircuit:
        pass

    @circuit.setter
    def circuit(self, circuit: QuantumCircuit) -> None:
        pass

    @property
    @abstractmethod
    def num_measurements(self) -> int:
        pass

    @num_measurements.setter
    def num_measurements(self, num_measurements: Optional[int]) -> None:
        pass

    @abstractmethod
    def execute(self) -> List[str]:
        pass



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 17, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from typing import Optional, Tuple

from ..platform import QuantumPlatform, QuantumProtocol
from .backend import CirqBackend
from .circuit import CirqCircuit
from .job import CirqJob


###############################################################################
## CIRQ PLATFORM
###############################################################################
class CirqPlatform(QuantumPlatform):
    def __init__(self) -> None:
        self.ERROR_MSG = f"{self.__class__.__name__}"  # TODO
        raise NotImplementedError(self.ERROR_MSG)

    ############################### PUBLIC API ###############################
    def create_circuit(self, num_qubits: int) -> CirqCircuit:
        raise NotImplementedError(self.ERROR_MSG)

    def create_job(  # type: ignore
        self,
        circuit: CirqCircuit,
        backend: CirqBackend,
        num_measurements: Optional[int] = None,
    ) -> CirqJob:
        raise NotImplementedError(self.ERROR_MSG)

    def fetch_random_bits(self, protocol: QuantumProtocol) -> str:
        raise NotImplementedError(self.ERROR_MSG)

    def retrieve_backend(
        self, max_bits_per_request: Optional[int] = None
    ) -> CirqBackend:
        raise NotImplementedError(self.ERROR_MSG)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qrand',
 'qrand.caches',
 'qrand.helpers',
 'qrand.platforms',
 'qrand.platforms.cirq',
 'qrand.platforms.qiskit',
 'qrand.platforms.qsharp',
 'qrand.protocols',
 'qrand.validation']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.4,<2.0.0', 'randomgen>=1.19.3,<2.0.0']

extras_require = \
{'cirq': ['cirq>=0.10.0,<0.11.0'],
 'qiskit': ['qiskit>=0.25.0,<0.26.0'],
 'qsharp': ['qsharp>=0.15.2103,<0.16.0']}

setup_kwargs = {
    'name': 'qrand',
    'version': '0.3.0a3',
    'description': 'A multiprotocol and multiplatform quantum random number generation framework',
    'long_description': '[![Unitary Fund](https://img.shields.io/badge/Supported_By-UNITARY_FUND-FFF000.svg?style=flat)](http://unitary.fund)\n[![YouTube](https://img.shields.io/badge/PR-qrand-FF0000.svg?style=flat&logo=YouTube&logoColor=white)](https://youtu.be/CG7BxuWFpME)\n[![PyPI](https://img.shields.io/pypi/v/qrand?label=PyPI&style=flat&color=3776AB&logo=Python&logoColor=white)](https://pypi.org/project/qrand/)\n[![Coverage](https://img.shields.io/badge/Coverage-67%25-green.svg?style=flat)](http://pytest.org)\n[![Apache-2.0 License](https://img.shields.io/github/license/pedrorrivero/qrand?label=License&style=flat&color=1D1D1D)](https://github.com/pedrorrivero/qrand/blob/master/LICENSE)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.4755731.svg)](https://doi.org/10.5281/zenodo.4755731)\n\n\n# qrand\n\n> A multiprotocol and multiplatform quantum random number generation framework\n\nRandom numbers are everywhere.\n\nComputer algorithms, data encryption, physical simulations, and even the arts use them all the time. There is one problem though: it turns out that they are actually very difficult to produce in large amounts. Classical computers can only implement mathematical tricks to emulate randomness, while measuring it out of physical processes turns out to be too slow. Luckily, the probabilistic nature of quantum computers makes these devices particularly useful for the task.\n\nQRAND is a free and open-source framework for quantum random number generation. Thanks to its loosely coupled design, it offers seamlessly compatibility between different [quantum computing platforms](#supported-quantum-platforms) and [QRNG protocols](#implemented-qrng-protocols). Not only that, but it also enables the creation of custom cross-compatible protocols, and a wide range of output formats (e.g. bitstring, int, float, complex, hex, base64).\n\nTo boost its efficiency, QRAND makes use of a concurrent cache to reduce the number of internet connections needed for random number generation; and for quality checks, it incorporates a suite of classical entropy validation tests which can be easily plugged into any base protocol.\n\nAdditionally, QRAND introduces an interface layer for [NumPy](https://numpy.org/) that enables the efficient production of quantum random numbers (QRN) adhering to a wide variety of probability distributions. This is ultimately accomplished by transforming uniform probability distributions produced in cloud-based real quantum hardware, through NumPy\'s random module.\n\n```python3\nfrom qrand import QuantumBitGenerator\nfrom qrand.platforms import QiskitPlatform\nfrom qrand.protocols import HadamardProtocol\nfrom numpy.random import Generator\nfrom qiskit import IBMQ\n\nprovider = IBMQ.load_account()\nplatform = QiskitPlatform(provider)\nprotocol = HadamardProtocol()\nbitgen = QuantumBitGenerator(platform, protocol)\ngen = Generator(bitgen)\n\nprint(f"Random Raw: {bitgen.random_raw()}")\nprint(f"Random Bitstring: {bitgen.random_bitstring()}")\nprint(f"Random Unsigned Int: {bitgen.random_uint()}")\nprint(f"Random Double: {bitgen.random_double()}")\n\nprint(f"Random Binomial: {gen.binomial(4, 1/4)}")\nprint(f"Random Exponential: {gen.exponential()}")\nprint(f"Random Logistic: {gen.logistic()}")\nprint(f"Random Poisson: {gen.poisson()}")\nprint(f"Random Std. Normal: {gen.standard_normal()}")\nprint(f"Random Triangular: {gen.triangular(-1, 0, 1)}")\n# ...\n```\n\n## Supported quantum platforms\nAs of May 2021, only [`Qiskit`](https://qiskit.org/) is supported. However, support for [`Cirq`](https://quantumai.google/cirq) and [`Q#`](https://docs.microsoft.com/en-us/azure/quantum/user-guide/?view=qsharp-preview) is under active development.\n\n## Implemented QRNG protocols\nAs of May 2021, only the basic `HadamardProtocol` is available. We are also working on implementing this [`EntaglementProtocol`](https://www.nature.com/articles/s41598-019-56706-2), as well as a version of [Google\'s Sycamore routine](https://arxiv.org/abs/1612.05903) (patent permitting).\n\n## Authors and citation\nQRAND is the work of many people who contribute to the project at\ndifferent levels. If you use QRAND, please cite as per the included\n[BibTeX file](QRAND.bib).\n\n<!-- ## Documentation -->\n\n## Contribution guidelines\nIf you\'d like to contribute to QRAND, please take a look at the\n[contribution guidelines](CONTRIBUTING.md). This project adheres to the following [code of conduct](CODE_OF_CONDUCT.md). By participating, you are expected to uphold this code.\n\nWe use [GitHub issues](https://github.com/pedrorrivero/qrand/issues) for tracking requests and bugs. Please use Unitary Fund\'s [Discord](http://discord.unitary.fund/) for discussion and simple questions.\n\n## Acknowledgements\nParts of this software\'s source code have been borrowed from the [qRNG](https://github.com/ozanerhansha/qRNG) project, which is licensed under the [GNU GPLv3](https://github.com/ozanerhansha/qRNG/blob/master/LICENSE) license. Copyright notice and specific changes can be found as a docstring wherever this applies.\n\n## License\n[Apache License 2.0](LICENSE)\n\n---\n(c) Copyright 2021 Pedro Rivero\n',
    'author': 'Pedro Rivero',
    'author_email': 'pedro.rivero.ramirez@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pedrorrivero/qrand',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: May 18, 2021
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2021 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from qiskit import QuantumCircuit as QiskitQuantumCircuit

from ..circuit import QuantumCircuit


###############################################################################
## QISKIT CIRCUIT
###############################################################################
class QiskitCircuit(QuantumCircuit, QiskitQuantumCircuit):
    def __init__(self, num_qubits: int) -> None:
        super(QuantumCircuit, self).__init__(num_qubits, num_qubits)

    @property
    def num_qubits(self) -> int:
        return super(QuantumCircuit, self).num_qubits

    ############################## SPECIAL GATES ##############################
    def measure(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).measure(target_qubit, target_qubit)

    ########################### SINGLE QUBIT GATES ###########################
    def h(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).h(target_qubit)

    def rx(self, radians: float, target_qubit: int) -> None:
        super(QuantumCircuit, self).rx(radians, target_qubit)

    def ry(self, radians: float, target_qubit: int) -> None:
        super(QuantumCircuit, self).ry(radians, target_qubit)

    def rz(self, radians: float, target_qubit: int) -> None:
        super(QuantumCircuit, self).rz(radians, target_qubit)

    def s(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).s(target_qubit)

    def t(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).t(target_qubit)

    def u1(self, theta: float, target_qubit: int) -> None:
        super(QuantumCircuit, self).u1(theta, target_qubit)

    def u2(self, phi: float, lam: float, target_qubit: int) -> None:
        super(QuantumCircuit, self).u2(phi, lam, target_qubit)

    def u3(
        self, theta: float, phi: float, lam: float, target_qubit: int
    ) -> None:
        super(QuantumCircuit, self).u3(theta, phi, lam, target_qubit)

    def x(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).x(target_qubit)

    def y(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).y(target_qubit)

    def z(self, target_qubit: int) -> None:
        super(QuantumCircuit, self).z(target_qubit)

    ############################# TWO QUBIT GATES #############################
    def cs(self, control_qubit: int, target_qubit: int) -> None:
        super(QuantumCircuit, self).cs(control_qubit, target_qubit)

    def cx(self, control_qubit: int, target_qubit: int) -> None:
        super(QuantumCircuit, self).cx(control_qubit, target_qubit)

    def cz(self, control_qubit: int, target_qubit: int) -> None:
        super(QuantumCircuit, self).cz(control_qubit, target_qubit)

    def swap(self, target_qubit_1: int, target_qubit_2: int) -> None:
        super(QuantumCircuit, self).swap(target_qubit_1, target_qubit_2)

    ############################ THREE QUBIT GATES ############################
    def ccx(
        self, control_qubit_1: int, control_qubit_2: int, target_qubit: int
    ) -> None:
        super(QuantumCircuit, self).ccx(
            control_qubit_1, control_qubit_2, target_qubit
        )

    def cswap(
        self, control_qubit: int, target_qubit_1: int, target_qubit_2: int
    ) -> None:
        super(QuantumCircuit, self).cswap(
            control_qubit, target_qubit_1, target_qubit_2
        )
